@@ -21,6 +21,7 @@
 
 import gettext
 
-__version__ = "2.1.4"
+__version__: str = "3.0.0"
+__reldate__: str = "May 29, 2024"
 
 gettext.install("duplicity", names=["ngettext"])
