@@ -30,26 +30,23 @@
 # any suggestions.
 
 
-import copy
+from dataclasses import dataclass
 import os
 import platform
 import resource
-import socket
 import sys
 import time
-from datetime import datetime
-
-import fasteners
+from textwrap import dedent
+from typing import Dict
 
 from duplicity import __version__
-from duplicity import asyncscheduler
+from duplicity import backend_pool
 from duplicity import cli_main
 from duplicity import config
 from duplicity import diffdir
 from duplicity import dup_collections
 from duplicity import dup_temp
 from duplicity import dup_time
-from duplicity import errors
 from duplicity import file_naming
 from duplicity import gpg
 from duplicity import log
@@ -67,7 +64,6 @@ exit_val = None
 
 def getpass_safe(message):
     import getpass
-    import locale
 
     return getpass.getpass(message)
 
@@ -319,31 +315,6 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
             end_block -= 1
         return start_index, start_block, end_index, end_block
 
-    def validate_block(orig_size, dest_filename):
-        info = backend.query_info([dest_filename])[dest_filename]
-        size = info["size"]
-        if size is None:
-            return  # error querying file
-        for attempt in range(1, config.num_retries + 1):
-            info = backend.query_info([dest_filename])[dest_filename]
-            size = info["size"]
-            if size == orig_size:
-                break
-            if size is None:
-                return
-            log.Notice(
-                _("%s Remote filesize %d for %s does not match local size %d, retrying.")
-                % (datetime.now(), size, util.escape(dest_filename), orig_size)
-            )
-            time.sleep(2**attempt)
-        if size != orig_size:
-            code_extra = f"{util.escape(dest_filename)} {int(orig_size)} {int(size)}"
-            log.FatalError(
-                _("File %s was corrupted during upload.") % os.fsdecode(dest_filename),
-                log.ErrorCode.volume_wrong_size,
-                code_extra,
-            )
-
     def put(tdp, dest_filename, vol_num):
         """
         Retrieve file size *before* calling backend.put(), which may (at least
@@ -353,7 +324,15 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
         putsize = tdp.getsize()
         if config.skip_volume != vol_num:  # for testing purposes only
             backend.put(tdp, dest_filename)
-        validate_block(putsize, dest_filename)
+        res, msg = backend.validate(dest_filename, putsize, source_path=tdp)
+        if not res:
+            code_extra = f"{util.escape(dest_filename)}: {msg}"
+            log.FatalError(
+                _("File %s was corrupted during upload.") % os.fsdecode(dest_filename),
+                log.ErrorCode.backend_validation_failed,
+                code_extra,
+            )
+
         if tdp.stat:
             tdp.delete()
         return putsize
@@ -388,6 +367,56 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
             fileobj = restore_get_enc_fileobj(config.backend, vol1_filename, manifest.volume_info_dict[1])
             fileobj.close()
 
+    @dataclass
+    class CommandMetaData:
+        vol_num: int
+        path_obj: path.Path
+        vol_info: manifest.VolumeInfo
+        transfer_success = False
+        manifest_written = False
+
+    def collect_put_results(bytes_written: int, backend_pooler, command2vol_map: Dict[int, CommandMetaData]) -> int:
+        commands_left, results = backend_pooler.collect_results()
+        for track_id, result in results.items():
+            size = result.result
+            bytes_written += size
+            progress.report_transfer(size, size)
+            log.Progress(_("Processed volume %d") % command2vol_map[track_id].vol_num, bytes_written)
+            command2vol_map[track_id].transfer_success = True
+            if command2vol_map[track_id].path_obj.stat:
+                command2vol_map[track_id].path_obj.delete()
+            if config.progress:
+                progress.tracker.snapshot_progress(command2vol_map[track_id].vol_num)
+            log.Debug(
+                f"Transfer of {command2vol_map[track_id].path_obj.get_filename()} with id {track_id} and size "
+                f"{size} took {result.get_runtime()}"
+            )
+        return commands_left
+
+    def write_manifest_in_sequence(mf, mf_file, command2vol_map: Dict[int, CommandMetaData]):
+        """
+        Ensure volume info is written only if the volume transfer was successful and in sequence
+        without gap otherwise missing volumes won't be detected and data is missing.
+        """
+        i = iter(command2vol_map.keys())
+        info = command2vol_map[next(i)]
+        while info.manifest_written and info.transfer_success:
+            # skip all entries already written
+            info = command2vol_map[next(i)]
+        while info.transfer_success:
+            # if next volumes are tranferred
+            if not info.manifest_written:
+                mf.add_volume_info(info.vol_info)
+            info.manifest_written = True
+            try:
+                info = command2vol_map[next(i)]
+            except StopIteration:
+                break
+            if info.vol_num == 1:
+                mf_file.to_partial()
+            else:
+                mf_file.flush()
+
     if not config.restart:
         # normal backup start
         vol_num = 0
@@ -420,18 +449,10 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
         progress.tracker.set_start_volume(vol_num + 1)
         progress.progress_thread.start()
 
-    # This assertion must be kept until we have solved the problem
-    # of concurrency at the backend level. Concurrency 1 is fine
-    # because the actual I/O concurrency on backends is limited to
-    # 1 as usual, but we are allowed to perform local CPU
-    # intensive tasks while that single upload is happening. This
-    # is an assert put in place to avoid someone accidentally
-    # enabling concurrency above 1, before adequate work has been
-    # done on the backends to make them support concurrency.
-    assert config.async_concurrency <= 1
-
-    io_scheduler = asyncscheduler.AsyncScheduler(config.async_concurrency)
-    async_waiters = []
+    backend_pooler = None
+    command2vol_map: Dict[int, CommandMetaData] = {}
+    if config.concurrency > 0:
+        backend_pooler = backend_pool.BackendPool(backend.backend.parsed_url.url_string, processes=config.concurrency)
 
     while not at_end:
         # set up iterator
@@ -455,31 +476,47 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
         vi = manifest.VolumeInfo()
         vi.set_info(vol_num, *get_indicies(tarblock_iter))
         vi.set_hash("SHA1", gpg.get_hash("SHA1", tdp))
-        mf.add_volume_info(vi)
 
         # Checkpoint after each volume so restart has a place to restart.
         # Note that until after the first volume, all files are temporary.
         if vol_num == 1:
             sig_outfp.to_partial()
-            man_outfp.to_partial()
         else:
             sig_outfp.flush()
-            man_outfp.flush()
 
         if config.skip_if_no_change and diffdir.stats.DeltaEntries == 0 and at_end and vol_num == 1:
             # if nothing changed, skip upload if configured.
-            msg = _(f"Skipped volume upload, as effectivly nothing has changed")
+            msg = _("Skipped volume upload, as effectivly nothing has changed")
             log.Progress(msg, diffdir.stats.SourceFileSize)
             log.Notice(_(msg))
             config.skipped_inc = True
             tdp.delete()
+            continue
+
+        if config.concurrency > 0 and backend_pooler:
+            try:
+                # use process pool the write volumes in the background.
+                progress.report_transfer(0, tdp.getsize())
+                track_id = backend_pooler.command(backend.put_validated.__name__, args=(tdp, dest_filename))
+                command2vol_map[track_id] = CommandMetaData(vol_num, tdp, vi)
+                commands_processing = collect_put_results(bytes_written, backend_pooler, command2vol_map)
+                log.Debug(f"Still {commands_processing} in the background pool.")
+                write_manifest_in_sequence(mf, man_outfp, command2vol_map)
+            except (Exception, SystemExit) as e:
+                # ensure pool processes terminate clean
+                backend_pooler.shutdown()
+                raise
+
         else:
-            # Write Volume to backend
-            async_waiters.append(
-                io_scheduler.schedule_task(
-                    lambda tdp, dest_filename, vol_num: put(tdp, dest_filename, vol_num), (tdp, dest_filename, vol_num)
-                )
-            )
+            # write Volume to backend - blocking
+            mf.add_volume_info(vi)
+
+            if vol_num == 1:
+                man_outfp.to_partial()
+            else:
+                man_outfp.flush()
+
+            bytes_written += put(tdp, dest_filename, vol_num)
 
             # Log human-readable version as well as raw numbers for machine consumers
             log.Progress(_("Processed volume %d") % vol_num, diffdir.stats.SourceFileSize)
@@ -488,13 +525,32 @@ def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
             if config.progress:
                 progress.tracker.snapshot_progress(vol_num)
 
-            # for testing purposes only - assert on inc or full
-            assert config.fail_on_volume != vol_num, f"Forced assertion for testing at volume {int(vol_num)}"
-
-    # Collect byte count from all asynchronous jobs; also implicitly waits
-    # for them all to complete.
-    for waiter in async_waiters:
-        bytes_written += waiter()
+    if backend_pooler:
+        try:
+            # wait for background commands, collect some stats and shutdown clean.
+            log.Debug("Collecting remaining results from backend pool.")
+            while True and backend_pooler:
+                commands_left = collect_put_results(bytes_written, backend_pooler, command2vol_map)
+                write_manifest_in_sequence(mf, man_outfp, command2vol_map)
+                if commands_left == 0:
+                    break
+                else:
+                    log.Debug(f"Still {commands_left} commands left, waiting ...")
+                    time.sleep(2)
+            # print some stats, that can be used to track efficiency of concurrent transfers.
+            stats = backend_pooler.get_stats()
+            log.Info(
+                f'Transferred {stats["count"]} volumes, a volume took avg: {stats["avg"]}, '
+                'max: {stats["max"]}, min: {stats["min"]}'
+            )
+            # add stats to jsonstat. Not the most elegant way, but it is working.
+            diffdir.stats.stat_attrs += ("VolumeTransferTimeAVG", "VolumeTransferTimeMAX", "VolumeTransferTimeMIN")
+            diffdir.stats.set_stat("VolumeTransferTimeAVG", stats["avg"])
+            diffdir.stats.set_stat("VolumeTransferTimeMAX", stats["max"])
+            diffdir.stats.set_stat("VolumeTransferTimeMIN", stats["min"])
+        finally:
+            # make sure pool get terminated gracefully in any case.
+            backend_pooler.shutdown()
 
     # Upload the collection summary.
     # bytes_written += write_manifest(mf, backup_type, backend)
@@ -1399,7 +1455,7 @@ def check_resources(action):
         # Calculate space we need for at least 2 volumes of full or inc
         # plus about 30% of one volume for the signature files.
         freespace = stats.f_frsize * stats.f_bavail
-        needspace = ((config.async_concurrency + 1) * config.volsize) + int(0.30 * config.volsize)
+        needspace = ((config.concurrency + 1) * config.volsize) + int(0.30 * config.volsize)
         if freespace < needspace:
             log.FatalError(
                 _("Temp space has %d available, backup needs approx %d.") % (freespace, needspace),
@@ -1522,14 +1578,16 @@ def main():
     # and refuse to run if it is set.
     if "PYTHONOPTIMIZE" in os.environ:
         log.FatalError(
-            _(
-                """\
-PYTHONOPTIMIZE in the environment causes duplicity to fail to
-recognize its own backups.  Please remove PYTHONOPTIMIZE from
-the environment and rerun the backup.
+            dedent(
+                _(
+                    """\
+                PYTHONOPTIMIZE in the environment causes duplicity to fail to
+                recognize its own backups.  Please remove PYTHONOPTIMIZE from
+                the environment and rerun the backup.
 
-See https://bugs.launchpad.net/duplicity/+bug/931175
-"""
+                See https://bugs.launchpad.net/duplicity/+bug/931175
+                """
+                )
             ),
             log.ErrorCode.pythonoptimize_set,
         )
@@ -1547,15 +1605,8 @@ See https://bugs.launchpad.net/duplicity/+bug/931175
     # determine what action we're performing and process command line
     action = cli_main.process_command_line(sys.argv[1:])
 
-    config.lockpath = os.path.join(config.archive_dir_path.name, b"lockfile")
-    config.lockfile = fasteners.process_lock.InterProcessLock(config.lockpath)
-    log.Debug(_("Acquiring lockfile %s") % os.fsdecode(config.lockpath))
-    if not config.lockfile.acquire(blocking=False):
-        log.FatalError(
-            "Another duplicity instance is already running with this archive directory\n", log.ErrorCode.user_error
-        )
-        log.shutdown()
-        sys.exit(2)
+    # make sure we have lock
+    util.acquire_lockfile()
 
     # log some status info
     log_startup_parms(log.INFO)
@@ -1597,7 +1648,7 @@ def do_backup(action):
                 break
             last_backup = last_full_chain.get_last()
             if last_backup.partial:
-                if action in ["backup", "full", "inc"]:
+                if action in ["full", "inc"]:
                     # set restart parms from last_backup info
                     config.restart = Restart(last_backup)
                     # (possibly) reset action
@@ -1630,7 +1681,7 @@ def do_backup(action):
         log.Notice(_("Last full backup date: none"))
     if (
         not config.restart
-        and action in ["inc", "backup"]
+        and action in ["inc"]
         and config.full_if_older_than is not None
         and last_full_time < dup_time.curtime - config.full_if_older_than
     ):
@@ -1669,7 +1720,7 @@ def do_backup(action):
     elif action == "sync":
         sync_archive(col_stats)
     else:
-        assert action in ["backup", "full", "inc"], action
+        assert action in ["full", "inc"], action
         # the passphrase for full and inc is used by --sign-key
         # the sign key can have a different passphrase than the encrypt
         # key, therefore request a passphrase
